@@ -25,8 +25,6 @@
 '''tables module specifying the database model'''
 
 import math
-import re
-from collections import OrderedDict
 from operator import attrgetter
 import numpy as np
 
@@ -36,23 +34,12 @@ from sqlalchemy.ext.associationproxy import association_proxy
 from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method
 
-import mendeleev
+from .econf import ElectronicConfiguration, get_l, ORBITALS
+
 
 __all__ = ['Element', 'IonizationEnergy', 'IonicRadius', 'OxidationState',
            'Isotope', 'Series', 'ScreeningConstant']
 
-ORBITALS = ('s', 'p', 'd', 'f', 'g', 'h', 'i', 'j', 'k')
-SHELLS = ('K', 'L', 'M', 'N', 'O', 'P', 'Q')
-
-
-def get_l(shell):
-    'Return the orbital angular momentum quantum number for a given shell'
-
-    if shell in ORBITALS:
-        return ORBITALS.index(shell.lower())
-    else:
-        raise ValueError('"{}" is not a proper shell label'.format(shell))
-
 
 Base = declarative_base()
 
@@ -72,12 +59,14 @@ class Element(Base):
         Atomic number
       atomic_radius : float
         Atomic radius in pm
+      atomic_radius_rahm : float
+        Atomic radius by Rahm et al. in pm
       atomic_volume : float
         Atomic volume in cm3/mol
       atomic_weight : float
         Relative atomic weight as the ratio of the average mass of atoms
         of the element to 1/12 of the mass of an atom of 12C
-      block : int
+      block : str
         Block in periodic table, s, p, d, f
       boiling_point : float
         Boiling temperature in K
@@ -89,6 +78,8 @@ class Element(Base):
       c6_gb : float
         C_6 dispersion coefficient in a.u. from Gould, T., & Bučko, T. (2016).
         JCTC, 12(8), 3603–3613. http://doi.org/10.1021/acs.jctc.6b00361
+      cas : str
+        Chemical Abstracts Service identifier
       covalent_radius_bragg : float
         Covalent radius in pm from
       covalent_radius_cordero : float
@@ -97,11 +88,14 @@ class Element(Base):
         Covalent radii revisited. Dalton Transactions, (21), 2832.
         doi:10.1039/b801115j
       covalent_radius_pyykko : float
-        Covalent radius in pm Pyykkö, P., & Atsumi, M. (2009). Molecular
-        Single-Bond Covalent Radii for Elements 1-118. Chemistry - A European
-        Journal, 15(1), 186–197. doi:10.1002/chem.200800987
-      covalent_radius_slater : float
-        Covalent radius in pm from
+        Single bond covalent radius in pm Pyykkö, P., & Atsumi, M. (2009).
+        Molecular Single-Bond Covalent Radii for Elements 1-118.
+        Chemistry - A European Journal, 15(1), 186–197.
+        doi:10.1002/chem.200800987
+      covalent_radius_pyykko_double : float
+        Double bond covalent radius in pm from P. Pyykkö et al.
+      covalent_radius_pyykko_triple : float
+        Triple bond covalent radius in pm from P. Pyykkö et al.
       cpk_color : str
         CPK color of the atom in HEX,
         see http://jmol.sourceforge.net/jscolors/#color_U
@@ -110,10 +104,15 @@ class Element(Base):
       description : str
         Short description of the element
       dipole_polarizability : float
-        Dipole polarizability in atomic units from P. Schwerdtfeger "Table of
-        experimental and calculated static dipole polarizabilities for the
-        electronic ground states of the neutral elements (in atomic units)",
-        February 11, 2014
+        Dipole polarizability in atomic units
+      dipole_polarizability_unc:  float
+        Uncertainty of the dipole polarizability
+      discoverers: str
+        The discoverers of the element
+      discovery_location: str
+        The location where the element was discovered
+      discovery_year: int
+        The year the element was discovered
       electron_affinity : float
         Electron affinity in eV
       en_allen : float
@@ -128,10 +127,18 @@ class Element(Base):
         Evaporation heat in kJ/mol
       fusion_heat : float
         Fusion heat in kJ/mol
-      gas_basicity : Float
+      gas_basicity : float
         Gas basicity
+      geochemical_class : str
+        Geochemical classification of the elements
+      glawe_number: int
+        Glawe number (scale)
+      goldschmidt_class : str
+        Goldschmidt classification of the elements
       group : int
-        Group in periodic table
+        Group number
+      group_id : Group
+        Group details
       heat_of_formation : float
         Heat of formation in kJ/mol
       is_monoisotopic : bool
@@ -146,24 +153,41 @@ class Element(Base):
       mass : float
         Relative atomic mass. Ratio of the average mass of atoms
         of the element to 1/12 of the mass of an atom of 12C
+      mendeleev_number : int
+        Mendeleev number
       melting_point : float
         Melting temperature in K
+      metallic_radius : Float
+        Single-bond metallic radius or metallic radius, have been
+        calculated by Pauling using interatomic distances and an
+        equation relating such distances with bond number
+      metallic_radius_c12 : Float
+        Metallic radius obtained by Pauling with an assumed number of
+        nearest neighbors equal to 12
       molcas_gv_color : str
         Color of an atom in HEX from MOLCAS GV http://www.molcas.org/GV/
       name : str
-        Name in english
+        Name in English
+      name_origin: str
+        Origin of the name
       period : int
         Period in periodic table
+      pettifor_number: int
+        Pettifor scale
       proton_affinity : Float
         Proton affinity
       series : int
         Index to chemical series
+      sources: str
+        Sources of the element
       specific_heat : float
         Specific heat in J/g mol @ 20 C
       symbol : str of length 1 or 2
         Chemical symbol
       thermal_conductivity : float
         Thermal conductivity in @/m K @25 C
+      uses: str
+        Uses of the element
       vdw_radius : float
         Van der Waals radius in pm from W. M. Haynes, Handbook of Chemistry and
         Physics 95th Edition, CRC Press, New York, 2014, ISBN-10: 1482208679,
@@ -196,21 +220,28 @@ class Element(Base):
     annotation = Column(String)
     atomic_number = Column(Integer, primary_key=True)
     atomic_radius = Column(Float)
+    atomic_radius_rahm = Column(Float)
     atomic_volume = Column(Float)
     atomic_weight = Column(Float)
     atomic_weight_uncertainty = Column(Float)
     block = Column(String)
     boiling_point = Column(Float)
+    cas = Column(String)
     covalent_radius_bragg = Column(Float)
     covalent_radius_cordero = Column(Float)
     covalent_radius_pyykko = Column(Float)
-    covalent_radius_slater = Column(Float)
+    covalent_radius_pyykko_double = Column(Float)
+    covalent_radius_pyykko_triple = Column(Float)
     c6 = Column(Float)
     c6_gb = Column(Float)
     cpk_color = Column(String)
     density = Column(Float)
     description = Column(String)
     dipole_polarizability = Column(Float)
+    dipole_polarizability_unc = Column(Float)
+    discoverers = Column(String)
+    discovery_location = Column(String)
+    discovery_year = Column(Integer)
     electron_affinity = Column(Float)
     en_allen = Column(Float)
     en_ghosh = Column(Float)
@@ -219,8 +250,11 @@ class Element(Base):
     evaporation_heat = Column(Float)
     fusion_heat = Column(Float)
     gas_basicity = Column(Float)
+    geochemical_class = Column(String)
+    glawe_number = Column(Integer)
+    goldschmidt_class = Column(String)
     group_id = Column(Integer, ForeignKey("groups.group_id"))
-    group = relationship("Group", uselist=False)
+    group = relationship("Group", uselist=False, lazy='subquery')
     heat_of_formation = Column(Float)
     is_monoisotopic = Column(Boolean)
     is_radioactive = Column(Boolean)
@@ -228,16 +262,23 @@ class Element(Base):
     lattice_constant = Column(Float)
     lattice_structure = Column(String)
     melting_point = Column(Float)
+    mendeleev_number = Column(Integer)
+    metallic_radius = Column(Float)
+    metallic_radius_c12 = Column(Float)
     molcas_gv_color = Column(String)
     name = Column(String)
+    name_origin = Column(String)
     period = Column(Integer)
+    pettifor_number = Column(Integer)
     proton_affinity = Column(Float)
     _series_id = Column("series_id", Integer, ForeignKey("series.id"))
-    _series = relationship("Series", uselist=False)
+    _series = relationship("Series", uselist=False, lazy='subquery')
     series = association_proxy("_series", "name")
+    sources = Column(String)
     specific_heat = Column(Float)
     symbol = Column(String)
     thermal_conductivity = Column(Float)
+    uses = Column(String)
     vdw_radius = Column(Float)
     vdw_radius_alvarez = Column(Float)
     vdw_radius_bondi = Column(Float)
@@ -248,11 +289,11 @@ class Element(Base):
     vdw_radius_uff = Column(Float)
     vdw_radius_mm3 = Column(Float)
 
-    ionic_radii = relationship("IonicRadius")
-    _ionization_energies = relationship("IonizationEnergy")
-    _oxidation_states = relationship("OxidationState")
-    isotopes = relationship("Isotope")
-    screening_constants = relationship('ScreeningConstant')
+    ionic_radii = relationship("IonicRadius", lazy='subquery')
+    _ionization_energies = relationship("IonizationEnergy", lazy='subquery')
+    _oxidation_states = relationship("OxidationState", lazy='subquery')
+    isotopes = relationship("Isotope", lazy='subquery')
+    screening_constants = relationship('ScreeningConstant', lazy='subquery')
 
     @reconstructor
     def init_on_load(self):
@@ -428,14 +469,14 @@ class Element(Base):
           o : str
             Orbital label, (s, p, d, ...)
           alle : bool
-            Use all the valence electrons, i.e. calculate screening for an extra
-            electron when method='slater', if method='clementi' this option is
-            ignored
+            Use all the valence electrons, i.e. calculate screening for an
+            extra electron when method='slater', if method='clementi' this
+            option is ignored
         '''
 
         # identify th valence s,p vs d,f
         if n is None:
-            n = self.ec.maxn()
+            n = self.ec.max_n()
         else:
             if not isinstance(n, int):
                 raise ValueError('<n> should be an integer, got: {}'.format(type(n)))
@@ -574,8 +615,8 @@ class Element(Base):
 
            \chi_{MB} = \sqrt{\\frac{1}{n_{v}}\sum^{n_{v}}_{k=1} I_{k}}
 
-        where: :math:`n_{v}` is the number of valence electrons and :math:`I_{k}`
-        is the :math:`k` th ionization potential.
+        where: :math:`n_{v}` is the number of valence electrons and
+        :math:`I_{k}` is the :math:`k` th ionization potential.
         '''
 
         ionenergies = [self.ionenergies.get(i, None)
@@ -606,15 +647,12 @@ class Element(Base):
         '''
 
         if charge is None or not isinstance(charge, int) or charge == 0:
-            raise ValueError('charge should be a nonzero  initeger')
+            raise ValueError('charge should be a nonzero initeger, got: {}'.format(charge))
 
         neff = {1: 0.85, 2: 1.99, 3: 2.89, 4: 3.45, 5: 3.85, 6: 4.36, 7: 4.99}
         RY = 13.605693009
 
-        if charge == 0:
-            Ie = self.electron_affinity
-        elif charge > 0:
-            Ie = self.ionenergies.get(charge, None)
+        Ie = self.ionenergies.get(charge, None)
 
         crs = [(IR.coordination, IR.spin, getattr(IR, radius))
                for IR in self.ionic_radii if IR.charge == charge]
@@ -622,7 +660,7 @@ class Element(Base):
         out = {}
         for coord, spin, cr in crs:
             # the 100.0 factor converts picometers to Angstroms
-            eneg = neff[self.ec.maxn()] * math.sqrt(Ie / RY) * 100.0 / cr
+            eneg = neff[self.ec.max_n()] * math.sqrt(Ie / RY) * 100.0 / cr
             if len(spin) < 1:
                 out[coord] = eneg
             else:
@@ -642,8 +680,10 @@ class Element(Base):
                 Radius to use in the calcualtion
         '''
 
+        from mendeleev.utils import estimate
+
         r = getattr(self, radius)
-        rng = mendeleev.interpolate(self.atomic_number, radius)
+        rng = estimate(self.atomic_number, radius)
 
         return math.pow(rng / r, 3)
 
@@ -665,7 +705,7 @@ class Element(Base):
 
 class IonicRadius(Base):
     '''
-    Effective ionic radii and crystal radii in pm retrieved from [1].
+    Effective ionic radii and crystal radii in pm retrieved from [1]_.
 
     .. [1] Shannon, R. D. (1976). Revised effective ionic radii and systematic
        studies of interatomic distances in halides and chalcogenides. Acta
@@ -742,7 +782,7 @@ class IonizationEnergy(Base):
 
     def __str__(self):
 
-        return "{1:5d} {2:10.5f}".format(self.degree, self.energy)
+        return "{0:5d} {1:10.5f}".format(self.degree, self.energy)
 
     def __repr__(self):
 
@@ -796,8 +836,9 @@ class Series(Base):
       name : str
         Name of the series
       color : str
-        The HEX representation of a color of the series, the colors we obtained
-        from `ColorBrewer <http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=10>`_
+        The HEX representation of a color of the series, the colors were
+        obtained from
+        `ColorBrewer <http://colorbrewer2.org/?type=qualitative&scheme=Paired&n=10>`_
         the qualitative 10-class paired colormap
     '''
 
@@ -853,13 +894,23 @@ class Isotope(Base):
 
     def __str__(self):
 
-        return "{0:5d} {1:5d} {2:10.5f}".format(
-            self.atomic_number, self.mass_number, self.mass)
+        afmt = '5.3f'
+        mfmt = '10.5f'
+
+        if self.mass is None:
+            mfmt = ''
+
+        if self.abundance is None:
+            afmt = ''
+
+        return "{0:5d} {1:5d} {2:{mfmt}} {3:{afmt}}".format(
+            self.atomic_number, self.mass_number, self.mass,
+            self.abundance, mfmt=mfmt, afmt=afmt)
 
     def __repr__(self):
 
-        return "<Isotope(Z={}, A={}, mass={})>".format(
-            self.atomic_number, self.mass_number, self.mass)
+        return "<Isotope(Z={}, A={}, mass={}, abundance={})>".format(
+            self.atomic_number, self.mass_number, self.mass, self.abundance)
 
 
 class ScreeningConstant(Base):
@@ -900,178 +951,3 @@ class ScreeningConstant(Base):
 
         return "<ScreeningConstant(Z={0:4d}, n={1:3d}, s={2:s}, screening={3:10.4f})>".format(
             self.atomic_number, self.n, self.s, self.screening)
-
-
-class ElectronicConfiguration(object):
-    '''Electronic configuration handler'''
-
-    def __init__(self, confstr, atomre=None, shellre=None):
-
-        self.noble = {
-            'He': '1s2',
-            'Ne': '1s2 2s2 2p6',
-            'Ar': '1s2 2s2 2p6 3s2 3p6',
-            'Kr': '1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6',
-            'Xe': '1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6 5s2 4d10 5p6',
-            'Rn': '1s2 2s2 2p6 3s2 3p6 4s2 3d10 4p6 5s2 4d10 5p6 6s2 4f14 5d10 6p6'
-        }
-
-        self.confstr = confstr
-        self.atomre = atomre
-        self.shellre = shellre
-
-        # parse the confstr and initialize core, valence and conf attributes
-        self.parse()
-
-    @property
-    def atomre(self):
-        'Regular expression for atomic symbols'
-        return self._atomre
-
-    @atomre.setter
-    def atomre(self, value):
-
-        if value is None:
-            self._atomre = re.compile(r'\[([A-Z][a-z]*)\]')
-        else:
-            self._atomre = re.compile(value)
-
-    @property
-    def shellre(self):
-        'Regular expression for the shell'
-        return self._shellre
-
-    @shellre.setter
-    def shellre(self, value):
-
-        if value is None:
-            self._shellre = re.compile(r'(?P<n>\d)(?P<o>[spdfghijk])(?P<e>\d+)?')
-        else:
-            self._shellre = re.compile(value)
-
-    def parse(self):
-        '''
-        Parse a string with electronic configuration into an OrderedDict
-        representation
-        '''
-
-        citems = self.confstr.split()
-
-        core = {}
-
-        if self.atomre.match(citems[0]):
-            symbol = str(self.atomre.match(citems[0]).group(1))
-            citems = citems[1:]
-            core = [self.shellre.match(o).group('n', 'o', 'e')
-                    for o in self.noble[symbol].split() if self.shellre.match(o)]
-        valence = [self.shellre.match(o).group('n', 'o', 'e')
-                   for o in citems if self.shellre.match(o)]
-
-        self.core = OrderedDict([((int(n), o), (int(e) if e is not None else 1)) for (n, o, e) in core])
-        self.valence = OrderedDict([((int(n), o), (int(e) if e is not None else 1)) for (n, o, e) in valence])
-        self.conf = OrderedDict(list(self.core.items()) + list(self.valence.items()))
-
-    def sort(self, inplace=True):
-
-        if inplace:
-            self.conf = OrderedDict(sorted(self.conf.items(),
-                                           key=lambda x: (x[0][0] +
-                                           get_l(x[0][1]), x[0][0])))
-        else:
-            return OrderedDict(sorted(self.conf.items(),
-                                      key=lambda x: (x[0][0] + get_l(x[0][1]),
-                                      x[0][0])))
-
-    def electrons_per_shell(self):
-
-        return {s: sum([v for k, v in self.conf.items() if k[0] == n])
-                for n, s in zip(range(1, self.maxn() + 1), SHELLS)}
-
-    def __repr__(self):
-
-        return self.conf2str(self.conf)
-
-    def __str__(self):
-
-        return self.conf2str(self.conf)
-
-    @staticmethod
-    def conf2str(dictlike):
-
-        return " ".join(["{n:d}{s:s}{e:d}".format(n=k[0], s=k[1], e=v)
-                         for k, v in dictlike.items()])
-
-    def shell2int(self):
-
-        return [(x[0], get_l(x[1]), x[2]) for x in self.conf]
-
-    def maxn(self):
-        'Return the largest value of pricncipal qunatum number for the atom'
-
-        return max([shell[0] for shell in self.conf.keys()])
-
-    def last_subshell(self, wrt='aufbau'):
-
-        if wrt == 'aufbau':
-            return sorted(self.conf.items(),
-                          key=lambda x: (x[0][0] + get_l(x[0][1]), x[0][0]))[-1]
-
-    def nvalence(self, block, method=None):
-        'Return the number of valence electrons'
-
-        if block in ['s', 'p']:
-            return sum([v for k, v in self.conf.items() if k[0] == self.maxn()])
-        elif block == 'd':
-            if method == 'simple':
-                return 2
-            else:
-                return self.conf[(self.maxn(), 's')] +\
-                    self.conf[(self.maxn() - 1, 'd')]
-        elif block == 'f':
-            return 2
-        else:
-            raise ValueError('wrong block: {}'.format(block))
-
-    def slater_screening(self, n, o, alle=False):
-        '''
-        Calculate the screening constant using the papproach introduced by
-        Slater in Slater, J. C. (1930). Atomic Shielding Constants. Physical
-        Review, 36(1), 57–64.
-        `doi:10.1103/PhysRev.36.57 <http://www.dx.doi.org/10.1103/PhysRev.36.57>`_
-
-        Args:
-          n : int
-            Principal quantum number
-          o : str
-            orbtial label, (s, p, d, ...)
-          alle : bool
-            Use all the valence electrons, i.e. calculate screening for
-            an extra electron
-        '''
-
-        if alle:
-            ne = 0
-        else:
-            ne = 1
-
-        if n == 1:
-            coeff = 0.3
-        else:
-            coeff = 0.35
-
-        if o in ['s', 'p']:
-            # get the number of valence electrons - 1
-            vale = float(sum([v for k, v in self.conf.items() if k[0] == n and k[1] in ['s', 'p']]) - ne)
-            n1 = sum([v * 0.85 for k, v in self.conf.items() if k[0] == n - 1])
-            n2 = sum([float(v) for k, v in self.conf.items() if k[0] in range(1, n - 1)])
-
-        elif o in ['d', 'f']:
-            # get the number of valence electrons - 1
-            vale = float(sum([v for k, v in self.conf.items() if k[0] == n and k[1] == o]) - ne)
-            n1 = sum([float(v) for k, v in self.conf.items() if k[0] == n and k[1] != o])
-            n2 = sum([float(v) for k, v in self.conf.items() if k[0] in range(1, n)])
-
-        else:
-            raise ValueError('wrong valence subshell: ', o)
-
-        return n1 + n2 + vale * coeff
